/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

  TODO
  parameterize for chunksize and nchunks
  add start-point arg to constructor
  lock in reader while cueingp is true?
  make locking in get() finer
  add locking to cueingp() etc
  mebbe p should be 0 based instead of datastart-based
cross-check:
(databytes/blockalign) == nsamples 
&& SamplesPerSec == (AvgBytesPerSec / BlockAlign) 
&& SamplesPerSec == (AvgBytesPerSec / waveformat.wChannels / bytespersamp)

  NOTES
  much is simplified if chunks contain an integral number of samps

*/

extern "C" {
#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
}

#include <string>
#include <algorithm>
#include <exception>
#include <cstddef>
#include <new>

#include <iostream>

using namespace std;

#include "wavfilea.h"
#include "dprintf.h"

float
wavfilea::size()
{
  pthread_mutex_lock(&m);

  float val = ((float) datalen) / (samplesize * nchannels * rate);

  pthread_mutex_unlock(&m);

  return val;
}

float
wavfilea::tell()
{
  pthread_mutex_lock(&m);

  float val = ((float) (p - datastart)) / (samplesize * nchannels * rate);

  pthread_mutex_unlock(&m);

  return val;
}

void
wavfilea::seek(float to)
{
  dprintf(1, "seek(%f)\n", to);

  off_t newp = f2off_t(to);

  dprintf(10, "newp = %ld\n", (long) newp);

  dprintf(10, "locking mutex in seek()\n");
  pthread_mutex_lock(&m);
  dprintf(10, "locked mutex in seek()\n");

  p = newp;

  // quick invalidation of pp by setting sent to 0
  // run_out_of_datap() will update pp if possible
  sent = 0;

  dprintf(10, "unlocking mutex in seek()\n");
  pthread_mutex_unlock(&m);
  dprintf(10, "unlocked mutex in seek()\n");
}

wavfilea::wavfilea(string name, size_t rate, size_t nchannels, bool quadraticp)
  : chunks(0), chunksize(default_chunksize), nchunks(default_nchunks), 
    file(name), fd(-1), phase(0), last(0, 0), last1(0, 0), last2(0, 0), 
    p(0), pp(0), sent(0),
    rate_wanted(rate), channels_wanted(nchannels), quadraticp(quadraticp),
    loopp(false), loopa(0), loopb(0),
    cueingpval(true), seekingpval(false), invalidpval(false)
{
  // allocate chunks
  // hmm..pretty evil..use vectors instead?
  chunks = reinterpret_cast<chunk *>(new char [sizeof(chunk) * nchunks]);
  for (size_t i = 0; i < nchunks; i++)
    new (&chunks[i]) chunk(chunksize);

  if ((fd = open(file.c_str(), O_RDONLY)) == -1)
    throw OpenError();

  // try to read and parse header
  unsigned char buf[1024];
  int n = read(fd, buf, sizeof(buf));

  if (n < 0)
    throw OpenError();
  
  parse_wavinfo(buf, n);

  pthread_mutex_init(&m, 0);
  pthread_cond_init(&readerc, 0);

  // init reader thread
  pthread_attr_t att;
  pthread_attr_init(&att);
  // make sure reader thread inherits our scheduling
  pthread_attr_setinheritsched(&att, PTHREAD_INHERIT_SCHED);
  if (pthread_create(&readert, &att, reader, this) == -1)
    {
      close(fd);
      pthread_attr_destroy(&att);
      throw exception();
    }
  pthread_attr_destroy(&att);
}

wavfilea::~wavfilea()
{
  pthread_cancel(readert);
  void *v;
  pthread_join(readert, &v);

  for (size_t i = 0; i < nchunks; i++)
    chunks[i].~chunk();

  delete [] reinterpret_cast<char *>(chunks);

  pthread_mutex_destroy(&m);
}  

static inline float
s16le2f(const unsigned char *p)
{
  unsigned char lsb = *p++;
  unsigned char msb = *p++;

  if (msb & 0x80)
    // a negative number
    {
      // implement the definition of a 16-bit two's complement number
      unsigned v = (((unsigned) msb) << 8) | ((unsigned) lsb);
      v -= 1;
      unsigned v1 = (~v) & 0xFFFF;
      float f = v1;
      f /= 32766;
      f *= -1;
      return f;
    }
  else
    {
      unsigned v = (((unsigned) msb) << 8) | ((unsigned) lsb);
      float f = v;
      f /= 32766;
      return f;
    }
}

// normalize a u8 to [-1.0, 1.0]
static inline float
u82f(unsigned char u)
{
  float f = u;
  f = (f - 128)/128;
  return f;
}

inline samppair
wavfilea::next_samp()
{
  samppair pr;

  if (samplesize == 2 && nchannels == 2)
    {
      pr.left = s16le2f(pp);
      pp += 2;
      p += 2;
      pr.right = s16le2f(pp);
      pp += 2;
      p += 2;
    }
  else if (samplesize == 2 && nchannels == 1)
    {
      pr.left = s16le2f(pp);
      pp += 2;
      p += 2;
      pr.right = pr.left;
    }
  else if (samplesize == 1 && nchannels == 2)
    {
      pr.left = u82f(*pp++);
      p += 1;
      pr.right = u82f(*pp++);
      p += 1;
    }
  else if (samplesize == 1 && nchannels == 1)
    {
      pr.left = u82f(*pp++);
      p += 1;
      pr.right = pr.left;
    }
  else
    {
      assert("can't happen");
      pr = samppair(0, 0);
    }

  return pr;
}

void
wavfilea::get(float *fbuf, size_t nsamps, float resample)
{
  dprintf(10, "locking mutex in get()\n");
  pthread_mutex_lock(&m);
  dprintf(10, "locked mutex in get()\n");

  // adjust resample by <file rate>/<rate wanted>
  resample *= (((float) rate)/((float) rate_wanted));

  if (loopp && p == loopb)
    p = loopa;

  dprintf(100, "quadraticp = %d in wavfilea::get(%s)\n", 
	  (int) quadraticp, file.c_str());

  if (run_out_of_datap() && (p < datastart + datalen))
    dprintf(1, "data underflow\n");

  if (quadraticp)
    // the linear and quadratic cases are virtually identical..only
    // two lines differ (they're marked in the quadratic case), but
    // common code is reproduced here for efficiency
    {
      for (size_t i = 0; i < nsamps; i++)
	{
	  // if we've run out of data (either at end-of-file or because of
	  // underflow), just discard the integer part of phase and use
	  // the old values of last and last1 below
	  if (run_out_of_datap())
	    phase -= floor(phase);
	  else
	    {
	      // chomp down any phase >= 1
	      for (; phase >= 1; phase--)
		{
		  if (run_out_of_datap())
		    // condition is handled the same as above
		    {
		      phase -= floor(phase);
		      break;
		    }
		  
		  // line not in linear case
		  last2 = last1;
		  last1 = last;
		  
		  // get the next sample pair
		  last = next_samp();
		  
		  // next_samp() is what increments p, so..
		  if (loopp && p == loopb)
		    p = loopa;
		}
	    }
	  
	  // phase < 1
	  
	  // line different in linear case
	  samppair pa = qinterp(phase+1, last2, last1, last);
	  
	  if (channels_wanted == 2)
	    {
	      *fbuf++ = pa.left;
	      *fbuf++ = pa.right;
	    }
	  else
	    *fbuf++ = (pa.left + pa.right)/2;
	  
	  phase += resample;
	}
    }
  else
    {
      for (size_t i = 0; i < nsamps; i++)
	{
	  // if we've run out of data (either at end-of-file or because of
	  // underflow), just discard the integer part of phase and use
	  // the old values of last and last1 below
	  if (run_out_of_datap())
	    phase -= floor(phase);
	  else
	    {
	      // chomp down any phase >= 1
	      for (; phase >= 1; phase--)
		{
		  if (run_out_of_datap())
		    // condition is handled the same as above
		    {
		      phase -= floor(phase);
		      break;
		    }
		  
		  last1 = last;
		  
		  // get the next sample pair
		  last = next_samp();
		  
		  // next_samp() is what increments p, so..
		  if (loopp && p == loopb)
		    p = loopa;
		}
	    }
	  
	  // phase < 1
	  
	  samppair pa = interp(phase, last1, last);
	  
	  if (channels_wanted == 2)
	    {
	      *fbuf++ = pa.left;
	      *fbuf++ = pa.right;
	    }
	  else
	    *fbuf++ = (pa.left + pa.right)/2;
	  
	  phase += resample;
	}
    }
    
  dprintf(10, "unlocking mutex in get()\n");
  pthread_mutex_unlock(&m);
  dprintf(10, "unlocked mutex in get\n");
}

static inline unsigned long
scoople32(const unsigned char *p)
{
  return 
    ((unsigned long) p[0])
    | (((unsigned long) p[1]) << 8)
    | (((unsigned long) p[2]) << 16)
    | (((unsigned long) p[3]) << 24);
}

static inline unsigned int
scoople16(const unsigned char *p)
{
  return 
    ((unsigned int) p[0])
    | (((unsigned int) p[1]) << 8);
}

// skip to the next RIFF chunk with the given magic in buffer p of length n
// return NULL if no such chunk found
// we assume we're at the beginning of *some* chunk
static inline const unsigned char *
skip2chunk(const char magic[4], const unsigned char *p, size_t n)
{
  while (true)
    {
      if (n < 8)		// no room for chunk header
	return 0;

      if (memcmp(magic, p, 4) == 0)
	return p;		// found!

      // otherwise, skip over chunk
      p += 4; n -= 4;
      unsigned long len = scoople32(p+4);
      p += 4; n -= 4;

      if (len > n)
	return 0;

      p += len;
      n -= len;
    }
}

void
wavfilea::parse_wavinfo(const unsigned char *p, int n)
{
  const unsigned char *porig = p;

  // could skip2chunk() for both RIFF and WAVE

  //  confirm `RIFF' header and skip size
  if (n < 8 || memcmp(p, "RIFF", 4) != 0)
    throw IllFormatted();
  p += 8;
  n -= 8;
  
  //  confirm `WAVE' header
  if (n < 4 || memcmp(p, "WAVE", 4) != 0)
    throw IllFormatted();
  p += 4;
  n -= 4;

  //  skip chunks to `fmt ' chunk
  const unsigned char *p1 = skip2chunk("fmt ", p, n);

  if (!p1)
    throw IllFormatted();

  n -= (p1 - p);
  p = p1;

  p += 4;			// skip past "fmt "
  n -= 4;

  if (n < 4)
    throw IllFormatted();

  unsigned long fmtlen = scoople32(p);
  p += 4;
  n -= 4;

  if (fmtlen < 16 || n < 16)
    throw IllFormatted();

  //  read in fields converting endianess
  int type = scoople16(p);
  p += 2;
  n -= 2;

  //  die if type isn't PCM
  if (type != 1)
    throw IllFormatted();

  nchannels = scoople16(p);
  p += 2;
  n -= 2;

  rate = scoople32(p);
  p += 4;
  n -= 4;

  //unsigned long avgbytespersec = scoople32(p);
  scoople32(p);
  p += 4;
  n -= 4;

  //  int blockalign = scoople16(p);
  scoople16(p);
  p += 2;
  n -= 2;

  //  read in bits/sample..make sure we can handle it
  int bitspersamp = scoople16(p);
  p += 2;
  n -= 2;

  if (bitspersamp == 8)
    {
      samplesize = 1;
      format = UNSIGNED;
    }
  else if (bitspersamp == 16)
    {
      samplesize = 2;
      format = SIGNED;
    }
  else
    throw IllFormatted();
    
  // skip any remaining fmt chunk data
  fmtlen -= 16;
  while (fmtlen > 0)
    {
      p++;
      n--;
      fmtlen--;
    }

  //  skip to `data' chunk and get it's size

  const unsigned char *p2 = skip2chunk("data", p, n);
  
  if (!p2)
    throw IllFormatted();

  n -= (p2 - p);
  p = p2;

  if (n < 8)
    throw IllFormatted();

  p += 4;
  n -= 4;
  unsigned long dsize = scoople32(p);
  p += 4;
  n -= 4;

  // OK..we've got everything and are at the data now
  datastart = p - porig;
  datalen = dsize;
}

bool
wavfilea::really_run_out_of_datap()
{
  // see if we just need to switch chunks
  for (size_t i = 0; i < nchunks; i++)
    {
      chunk *c = &chunks[i];

      if (c->validp && p >= c->off && p < c->off + c->dlen)
	{
	  pp = c->p + (p - c->off);
	  sent = c->p + c->dlen;
	  // let the reader know we've crossed chunks
	  pthread_cond_signal(&readerc);
	  return false;
	}
    }

  //dprintf(1, "data underflow!\n");
  // otherwise we've run out of data (maybe due to a seek)
  // let the reader know we need help..but only if we're not at eod
  if (p < datastart + datalen)
    pthread_cond_signal(&readerc);
  return true;
}

// convert a time in seconds to an offset in the wav file
inline off_t
wavfilea::f2off_t(float a)
{
  off_t ap;

  if (a < 0)
    ap = datastart;
  else
    {
      size_t off =  static_cast<size_t>(a * (nchannels * samplesize * rate));

      // normalize off to a sample boundary
      int alignment = nchannels * samplesize;
      off = (off / alignment) * alignment;

      if (off > datalen)
	ap = datastart + datalen;
      else
	ap = datastart + off;
    }

  return ap;
}

void
wavfilea::setloop(float a, float b)
{
  pthread_mutex_lock(&m);

  loopp = true;
  loopa = f2off_t(a);
  loopb = f2off_t(b);

  pthread_mutex_unlock(&m);
}

bool
wavfilea::getloop(float &a, float &b)
{
  pthread_mutex_lock(&m);

  if (loopp)
    {
      a = loopa;
      b = loopb;
    }

  bool val = loopp;

  pthread_mutex_unlock(&m);

  return val;
}

void
wavfilea::clearloop()
{
  pthread_mutex_lock(&m);

  loopp = false;

  pthread_mutex_unlock(&m);
}

/* figure out the next chunk we should read (if any) */
// this needs to give a chunk num, file pos and read length)
//    there's at least one valid chunk (cause of the way reader() inited
// NB: this interface is designed so scheduling can get clever when 
//     loops are handled by these players
bool
wavfilea::schedule_chunk(int &cn, off_t &fpos, off_t &rlen)
{
  dprintf(10, "locking mutex in schedule\n");
  pthread_mutex_lock(&m);
  dprintf(10, "locked mutex in schedule\n");

  dprintf(10, "p == %ld in schedule\n", (long) p);

  // don't do anything if p is past eod
  if (p >= datastart + datalen)
    {
      dprintf(10, "unlocking mutex in schedule\n");
      pthread_mutex_unlock(&m);
      dprintf(10, "unlocked mutex in schedule\n");
      
      return false;
    }

  bool retval = false;

  // offset of start of any chunk p would be in
  off_t pchunkoff = (((p - datastart) / chunksize) * chunksize) + datastart;

  // offset of start of any chunk a loop head would be in
  off_t lchunkoff;
  if (loopp)
    lchunkoff = (((loopa - datastart) / chunksize) * chunksize) + datastart;

  bool invalidp = false;	// true if there's an invalid chunk
  size_t invalidn;		// the # of an invalid chunk if any
  off_t maxchunkoff = 0;
  off_t minchunkoff = datastart + datalen;
  size_t maxchunkn;
  size_t minchunkn;
  bool pchunkp = false;
  size_t pchunkn;		// # of a chunk containing p
  bool p1chunkp = false;
  size_t p1chunkn;		// # of the chunk after p
  bool p_1chunkp = false;
  size_t p_1chunkn;		// # of the chunk b4 p
  bool lchunkp = false;
  size_t lchunkn;		// # of chunk containing loop head
  bool l1chunkp = false;
  size_t l1chunkn;		// # of chunk after loop head

  // index of a chunk not in core set..should exist if we have at least 
  //  6 chunks
  // might want to expand core by a chunk on each side though
  size_t noncoren;		

  // chunk analysis loop
  //
  // see if there are any invalid chunks and choose one if there are
  // calculate the min and max chunk file offsets and which chunks have them
  for (size_t i = 0; i < nchunks; i++)
    {
      if (!chunks[i].validp)
	{
	  invalidp = true;
	  invalidn = i;
	}
      else
	{
	  if (chunks[i].off >= maxchunkoff)
	    {
	      maxchunkoff = chunks[i].off;
	      maxchunkn = i;
	    }
	  if (chunks[i].off < minchunkoff)
	    {
	      minchunkoff = chunks[i].off;
	      minchunkn = i;
	    }

	  if (chunks[i].off == pchunkoff)
	    // the chunk containing p
	    {
	      pchunkp = true;
	      pchunkn = i;
	    }
	  // NB: the inequality ensures the subtraction is valid
	  else if (chunks[i].off < pchunkoff 
		   && chunks[i].off == pchunkoff - chunksize)
	    // the chunk b4 that containing p
	    {
	      p_1chunkp = true;
	      p_1chunkn = i;
	    }
	  // NB: the inequality ensures the addition is valid
	  else if (chunks[i].off > pchunkoff
		   && chunks[i].off == pchunkoff + chunksize)
	    // the chunk after that containing p
	    {
	      p1chunkp = true;
	      p1chunkn = i;
	    }

	  // SUBTLETY ALERT!
	  // note these may be the same as one of the preceeding 3
	  // chunks..that's why this isn't an "else if"
	  if (loopp && chunks[i].off == lchunkoff)
	    {
	      lchunkp = true;
	      lchunkn = i;
	    }
	  else if (loopp 
		   && chunks[i].off > lchunkoff 
		   && chunks[i].off == lchunkoff + chunksize)
	    {
	      l1chunkp = true;
	      l1chunkn = i;
	    }
	}
    }

  // separate loop to select noncoren due to above noted subtlety
  for (size_t i = 0; i < nchunks; i++)
    if (i != pchunkn
	&& i != p1chunkn
	&& i != p_1chunkn
	&& i != lchunkn
	&& i != l1chunkn)
      noncoren = i;
  
  // ensure we have our core set in a good priority order
  if (!pchunkp)
    {
      fpos = pchunkoff;
      retval = true;
    }
  else if (!p1chunkp && (datastart + datalen > pchunkoff + chunksize))
    {
      fpos = pchunkoff + chunksize;
      retval = true;
    }
  else if (!p_1chunkp && pchunkoff != datastart)
    {
      fpos = pchunkoff - chunksize;
      retval = true;
    }
  else if (loopp && !lchunkp && lchunkoff < datastart + datalen)
    {
      fpos = lchunkoff;
      retval = true;
    }
  else if (loopp && !l1chunkp && lchunkoff + chunksize < datastart + datalen)
    {
      fpos = lchunkoff + chunksize;
      retval = true;
    }

  if (retval)
    // if we're scheduling a chunk
    {
      // choose the victim
      // choose either an invalid or noncore chunk in that order 
      // of preference
      if (invalidp)
	cn = invalidn;
      else
	cn = noncoren;

      // make sure the victim chunk is marked invalid
      chunks[cn].validp = false;

      // ..and set rlen
      rlen = min<off_t>(chunksize, (datastart + datalen) - fpos);
    }

  // if core set is full, we'll return false

  dprintf(10, "unlocking mutex in schedule\n");
  pthread_mutex_unlock(&m);
  dprintf(10, "unlocked mutex in schedule\n");
  return retval;
}

void *
wavfilea::reader(void *p)
{
  wavfilea *w = static_cast<wavfilea *>(p);

  // read first chunk of data
  if (lseek(w->fd, w->datastart, SEEK_SET) == -1)
    cerr << "seek failed! " << strerror(errno) << endl;	// what to do?

  ssize_t n1 = min<size_t>(w->chunksize, w->datalen);

  ssize_t cnt = read(w->fd, w->chunks[0].p, n1);

  if (cnt != n1)
    cerr << "only read " << cnt << "bytes!" << endl; // what to do?
  else
    {
      w->chunks[0].off = w->datastart;
      w->chunks[0].dlen = cnt;
      w->chunks[0].validp = true;

      // init state for get()
      w->pp = w->chunks[0].p;
      w->p = w->datastart;
      w->pp = w->chunks[0].p;
      w->sent = w->chunks[0].p + w->chunks[0].dlen;
    }

  // indicate that we're ready to play
  w->cueingpval = false;

  while (true)
    {
      int cn;
      off_t fpos;
      off_t rlen;

      if (w->schedule_chunk(cn, fpos, rlen))
	{
	  dprintf(10, "scheduler returned true with %ld %ld %ld\n",
		  (long) cn, (long) fpos, (long) rlen);
	  //  initiate read
	  if (lseek(w->fd, fpos, SEEK_SET) == -1)
	    // what to do?
	    dprintf(1, "seek failed: %s\n", strerror(errno));
	  else
	    {
	      ssize_t cnt = read(w->fd, w->chunks[cn].p, rlen);
	      if (cnt != rlen)
		// what to do?
		dprintf(1, "read() read only %ld bytes: %s\n", 
			(long) cnt, strerror(errno));
	      else
		{
		  dprintf(10, "locking mutex after read()\n");
		  pthread_mutex_lock(&w->m);
		  dprintf(10, "locked mutex after read()\n");
		  w->chunks[cn].off = fpos;
		  w->chunks[cn].dlen = rlen;
		  w->chunks[cn].validp = true;
		  dprintf(10, "unlocking mutex after validating chunk\n");
		  pthread_mutex_unlock(&w->m);
		  dprintf(10, "unlocked mutex after validating chunk\n");
		}
	    }
	}
      else
	{
	  // wait until the consumer thinks we should try scheduling again
	  dprintf(10, "locking b4 wait\n");
	  pthread_mutex_lock(&w->m);
	  dprintf(10, "locked b4 wait\n");
	  pthread_cond_wait(&w->readerc, &w->m);
	  dprintf(10, "unlocking mutex after wait\n");
	  pthread_mutex_unlock(&w->m);
	  dprintf(10, "unlocked mutex after wait\n");
	}
    }
}

//#define TEST
#ifdef TEST

#include <cstdlib>

extern "C"
{
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <sched.h>
#include <fcntl.h>
#include <stdlib.h>
}

// test app to just open a file and write it to /dev/dsp at a 
//  given resampling rate

void
fv2u8v(float *a, unsigned char *u, size_t n)
{
  for (size_t i = 0; i < n; i++)
    u[i] = static_cast<unsigned char>((a[i] * 128) + 128);
}

static inline short
f2s16le(float f)
{
  return (short) (f * 32766);
}

static void
convert2(float *in, size_t nsamps, unsigned char *out)
{
  for (size_t i = 0; i < nsamps * 2; i++)
    {
      short v1 = f2s16le(*in++);
      unsigned short v2 = (unsigned short) v1;
      *out++ = (v2 & 0xFF);
      *out++ = (v2 >> 8);
    }
}

int
main(int argc, char **argv)
{
  if (argc != 3)
    return 1;

  if (geteuid() == 0)
    {
      struct sched_param sp;
      sp.sched_priority = 50;
      if (sched_setscheduler(0, SCHED_RR, &sp) == -1)
	{
	  dprintf(1, "couldn't set real-time scheduling: %s\n",
		  strerror(errno));
	  return 1;
	}
    }

  wavfilea *w1 = new wavfilea(argv[1], 44100, 2);
  delete w1;

  wavfilea w(argv[1], 44100, 2);
  float rate = atof(argv[2]);

  int dspfd = open("/dev/dsp", O_WRONLY);

  if (dspfd == -1)
    return 1;

  // whoo!  go hard real-time!
  // this first def is around 1/57th second latency
  int fragarg = 0x00060009;
  // this second one is 1/25th second latency
  //int fragarg = 0x0007000A;

  if (ioctl(dspfd, SNDCTL_DSP_SETFRAGMENT, &fragarg) == -1)
    return 1;

  int blocksize;
  if (ioctl(dspfd, SNDCTL_DSP_GETBLKSIZE, &blocksize) == -1)
    return 1;

  int bits = 16;

  if (ioctl(dspfd, SNDCTL_DSP_SAMPLESIZE, &bits) == -1)
    return 1;

  int stereop = 1;

  if (ioctl(dspfd, SNDCTL_DSP_STEREO, &stereop) == -1)
    return 1;

  unsigned long ulrate = 44100;

  if (ioctl(dspfd, SNDCTL_DSP_SPEED, &ulrate) == -1)
    return 1;

  // sigh..this'll need fencepost checkin
  float *a = new float[blocksize/2];
  unsigned char *u = new unsigned char[blocksize];

  // just busy/wait til the player is ready
  while (!(w.goodtogetp()))
    ;

  while (true)
    {
      w.get(a, blocksize/4, rate);
      convert2(a, blocksize/4, u);
      write(dspfd, u, blocksize);
    }

  return 0;
}
#endif
